#include cmlmt.sdf

proc (0) = EchoSSM(Z,d,H,T,c,R,Q,a_0,P_0);
@call SSM(Z,d,H,T,c,R,Q,0);@
"******************************";
"*    Measurement Equation    *";
"******************************";
format /rds 0,0;
? "y(t) =    Z(t)    x  alpha(t)  +    d(t)    +    epsilon(t)";
? "     =    " rows(Z) "x " cols(Z) "     " rows(A_0) "x " cols(a_0) "        " rows(d) "x 1        " rows(H) "x 1";
format /mb1 /ros 16,8;
"*    Z matrix    *";
Z;
"*    d vector    *";
d;
"*    H matrix    *";
H;
"*****************************";
"*    Transition Equation    *";
"*****************************";
format /rds 0,0;
? "alpha(t) =    T(t)    x alpha(t-1) +    c(t)    +    R(t)    x    nu(t)";
? "              " rows(T) "x " cols(T) "    " rows(A_0) "x " cols(a_0)  "          " rows(c) "x 1        " rows(R) "x " Cols(R);
format /mb1 /ros 16,8;
"*    T matrix    *";;
T;
"*    c vector    *";;
c;
"*    R matrix    *";;
R;
"*    Q matrix    *";;
Q;
"****************************";
"*    Initial Conditions    *";
"****************************";
"*    A_0 matrix    *";
a_0';
"*    P_0 matrix    *";;
P_0;

endp;

@***********************************@
@ proc to Update Initial Conditions @
@***********************************@
proc InitP0(T,R,Q); @ Just pass the stationary components of T,R and Q @
  local P_0;
  local m,g,oldtrap,w1,w2,w3;

  m = rows(T);
  g = rows(Q);
  
  if Q == zeros(g,g);
    P_0 = zeros(m,m);
    retp(P_0);
  else;
    w1 = T.*.T;
    w2 = eye(m^2) - w1;

    oldtrap = trapchk(1);
    trap 1,1;
    w2 = inv(w2);
    trap oldtrap,1;

    if scalerr(w2);      errorlog "error: The state space model is not stable.";      retp(error(0));    endif;

    w3 = w2*vec(R*Q*R');
    P_0 = reshape(w3,m,m); P_0 = P_0';
  endif;

  retp(P_0);
endp;

@**************************************************@
@ Find the last valid observation for a given date @
@**************************************************@
proc findlobs(series);
local lastobs, i;

lastobs=rows(series);
for i (rows(series), 1, -1);
    if not scalmiss(series[i]);
        lastobs=i;
        break;
    endif;
endfor;
retp(lastobs);
endp;

@***************************************************@
@ Find the first valid observation for a given date @
@                                              ==== @
@***************************************************@
proc findfobs(series);
local startobs, i;

startobs=0;
for i (1, cols(series), 1);
    if not scalmiss(series[i]);
        startobs=i;
        break;
    endif;
endfor;
retp(startobs);
endp;

@****************************************************************************@
@ Retrieve the first n valid vintages for a given period from a RT data set. @
@                                     ==================                     @
@****************************************************************************@
proc getvints(series,n);
local startobs, i, j, newdata, nvints, nobs;

nvints = cols(series);// # of vintages
nobs = rows(series);// max # of observations
newdata = miss(zeros(nvints,n),0); // set up an nvints-n+1 by n matrix of missing values

    for i (1, nvints, 1);
        j = findlobs(series[.,i]);
        if i+n-1 > nvints;
            // newdata[i,1:nvints-i+1] = series[j,i:nvints];    
            break;
        else;            
            newdata[i,1:n] = series[j,i:i+n-1];    
        endif;
    endfor;
    newdata=packr(newdata);

retp(newdata);
endp;

@****************************************@
@ proc to set up the system matrices for @
@  the Harvey-Clark model                @
@****************************************@
proc llfHarveyClark(struct PV pv1, struct DS ds1, ind);
    local H,Q,Z,d,T,c,R,a0,P0,Qscale,y;
    
    T = pvUnpack(pv1,"T");
    Q = pvUnpack(pv1,"Q"); Q = Q*Q';
    Z = ds1[2].dataMatrix;
    d = ds1[3].dataMatrix;
    H = ds1[4].dataMatrix; H = H*H';
    c = ds1[5].dataMatrix;
    R = ds1[6].dataMatrix;
    Qscale = ds1[9].dataMatrix; 
        Q = Q*diagrv(eye(rows(Q)),10^Qscale);
    call SSM_build(Z,d,H,T,c,R,Q,0);

    y = ds1[1].dataMatrix;

    a0 = ds1[7].dataMatrix;
    P0 = ds1[8].dataMatrix; P0 = P0*P0';
    call KFiltering(y,a0,P0);

    struct modelResults mm;
    if ind[1];
        mm.function = KF_ml;
    elseif rows(ind) > 3;
        if ind[4] == 1;
            mm.function = KF_matrix(3);
        elseif ind[4] == 2;
            mm.function = KF_matrix(4);
        elseif ind[4] == 3;
            {mm.function,Qscale} = Ksmoothing;
        elseif ind[4] == 4;
            {Qscale,mm.function} = Ksmoothing;
        endif;
    endif;
    retp(mm);
endp;

@**************************************************@
@ proc to set up the system matrices for           @
@  the Pure Noise model with AR(p) "true" dynamics @
@**************************************************@
proc llfARNoise(struct PV pv1, struct DS ds1, ind);
    local H,Q,Z,d,T,c,R,a0,P0,Qscale,Tscale,y;
    
    T = pvUnpack(pv1,"T");
    R = pvUnpack(pv1,"R");
    Z = ds1[2].dataMatrix;
    d = ds1[3].dataMatrix;
    H = ds1[4].dataMatrix; H = H*H';
    c = ds1[5].dataMatrix;
    Q = ds1[6].dataMatrix;
    Tscale = ds1[8].dataMatrix; 
    Qscale = ds1[9].dataMatrix; 
    R = R*diagrv(eye(cols(R)),10^Qscale);
    if rows(Tscale)==m;  T[p+1:p+m,p+1:p+m]=T[p+1:p+m,p+1:p+m]*diagrv(eye(m),10^Tscale);     endif;
    call SSM_build(Z,d,H,T,c,R,Q,0);

    y = ds1[1].dataMatrix;
    a0 = ds1[7].dataMatrix;
    P0 = InitP0(T,R,Q);
    call KFiltering(y,a0,P0);

    struct modelResults mm;
    if ind[1];
        mm.function = KF_ml;
    elseif rows(ind) > 3;
        if ind[4] == 1;
            mm.function = KF_matrix(3);
        elseif ind[4] == 2;
            mm.function = KF_matrix(4);
        elseif ind[4] == 3;
            {mm.function,Qscale} = Ksmoothing;
        elseif ind[4] == 4;
            {Qscale,mm.function} = Ksmoothing;
        endif;
    endif;
    retp(mm);
endp;

@*************************************************@
@ proc to set up the system matrices for          @
@  the Pure News model with AR(p) "true" dynamics @
@*************************************************@
// Unlike ARNoise, the R matrix passed is now a m+1 vector;
proc llfARNews(struct PV pv1, struct DS ds1, ind);
    local H,Q,Z,d,T,c,R,a0,P0,Qscale,Tscale,y,R1,m,p;
    
    T = pvUnpack(pv1,"T");
    R1 = pvUnpack(pv1,"R");
    Z = ds1[2].dataMatrix;
    d = ds1[3].dataMatrix;
    H = ds1[4].dataMatrix; H = H*H';
    c = ds1[5].dataMatrix;
    Q = ds1[6].dataMatrix;
    Tscale = ds1[8].dataMatrix; Qscale = ds1[9].dataMatrix; 
    R = ds1[10].dataMatrix; 
    p = ds1[11].dataMatrix; 
    m = ds1[12].dataMatrix; 
            
    R1 = diagrv(eye(m+1),10^Qscale)*R1;
    R[1,1:m+1] = R1[1:m+1,1]'; // Shocks to true values and "news"
    R[1+p:p+m,2:m+1] = -upmat(ones(m,1)*R1[2:m+1,1]');
    if rows(Tscale)==m;  T[p+1:p+m,p+1:p+m]=T[p+1:p+m,p+1:p+m]*diagrv(eye(m),10^Tscale);     endif;
    
    call SSM_build(Z,d,H,T,c,R,Q,0);
    y = ds1[1].dataMatrix;

    a0 = ds1[7].dataMatrix;
//    P0 = ds1[8].dataMatrix; P0 = P0*P0';
    P0 = InitP0(T,R,Q);

//    EchoSSM(Z,d,H,T,c,R,Q,a0,P0); // Useful for debugging

    call KFiltering(y,a0,P0);

    struct modelResults mm;
    if ind[1];
        mm.function = KF_ml;
    elseif rows(ind) > 3;
        if ind[4] == 1;
            mm.function = KF_matrix(3);
        elseif ind[4] == 2;
            mm.function = KF_matrix(4);
        elseif ind[4] == 3;
            {mm.function,Qscale} = Ksmoothing;
        elseif ind[4] == 4;
            {Qscale,mm.function} = Ksmoothing;
        endif;
    endif;
    retp(mm);
endp;

@*****************************************************@
@ proc to set up the system matrices for Harvey-Clark @
@*****************************************************@
proc (0) = setHarveyClark;
    TSM_Z = ones(1,2)~zeros(1,2);    TSM_d = 0;
    TSM_H = 0; @Does this have to be 4x4? @
    TSM_T = zeros(4,4); TSM_T[1,1 4]=1~1; TSM_T[3,2]=1; TSM_T[4,4]=1;
    TSM_c = zeros(4,1); @ Does this have to be 4x1? @
    TSM_R = eye(4); TSM_R[3,3] = 0;
    TSM_Q = zeros(4,4);
    TSM_a_0 = zeros(4,1);
    TSM_P_0 = eye(4);
endp;

@*********************************************************@
@ proc to set up the system matrices for                  @
@  the Pure Noise model with AR(p) "true" dynamics        @
@*********************************************************@
proc (0) = setARNoise(m,p);
// m is the row dimension of y(t) = # of vintages in model
// p is the order of the AR dynamics; p > 0 required
    ?;"setARNoise: Setting up system matrices for Pure Noise model with ";; 
        m;; " vintages and ";; p;; "th order AR dynamics.";
    if      m < 1; "ERROR: m > 0 required! Aborting....";
    elseif  p < 1; "ERROR: p > 0 required! Aborting....";
    else;
        TSM_Z = ones(m,1)~zeros(m,p-1)~eye(m);
        TSM_d = zeros(m,1);
        TSM_H = zeros(m,m);
        TSM_T = zeros(m+p,m+p);
            if p>1;
                TSM_T[2:p,1:p-1]=eye(p-1); 
            endif;
        TSM_c = zeros(m+p,1);
        TSM_R = zeros(m+p,m+1); 
        TSM_Q = eye(m+1);
        TSM_a_0 = zeros(m+p,1);
        TSM_P_0 = eye(m+p);
    endif;
endp;

@*********************************************************@
@ proc to set up the system matrices for                  @
@  the Pure News model with AR(p) "true" dynamics         @
@*********************************************************@
proc (0) = setARNews(m,p);
// m is the row dimension of y(t) = # of vintages in model
// p is the order of the AR dynamics; p > 0 required
    ?;"setARNews: Setting up system matrices for Pure Noise model with ";; 
        m;; " vintages and ";; p;; "th order AR dynamics.";
    if      m < 1; "ERROR: m > 0 required! Aborting....";
    elseif  p < 1; "ERROR: p > 0 required! Aborting....";
    else;
        TSM_d = zeros(m,1);
        TSM_H = zeros(m,m);
        TSM_T = zeros(m+p,m+p);
            if p>1;
                TSM_T[2:p,1:p-1]=eye(p-1); 
                TSM_Z = ones(m,1)~zeros(m,p-1)~eye(m);
            else;
                TSM_Z = ones(m,1)~eye(m);
            endif;
        TSM_c = zeros(m+p,1);
        TSM_R = zeros(m+p,m+1); 
        TSM_Q = eye(m+1);
        TSM_a_0 = zeros(m+p,1);
        TSM_P_0 = eye(m+p);
    endif;
endp;

