#include cmlmt.sdf

@*********************************************************@
@ proc to set up the system matrices for                  @
@  the News + Noise model with AR(p) "true" dynamics      @
@*********************************************************@
proc (2) = setARNewsNoise(x,p);
//
// x is the data matrix to analyse:
//      one vintage per column, no missing values)
// p is the order of the AR dynamics; p > 0 required
//
// The proc returns two initialized structures {PV,DS}
//
// The DS structure has the form
//      DS[1] = y (data to estimate)
//      DS[2] = Z
//      DS[3] = d
//      DS[4] = H
//      DS[5] = c
//      DS[6] = Q 
//      DS[7] = a_0
//      DS[8] = P_0 (not used)
//      DS[9] = NsScale or Qscale
//      DS[10] = NwScale
//      DS[11] = m
//      DS[12] = p
//
//  The PV structure has the form
//      PV[1] = T (masked)
//      PV[2] = RNoise (mx1 vector)
//      PV[3] = RNews  (mx1 vector)
//      PV[4] = ARSigma  (scalar)
//

local m, TSM_d, TSM_H, TSM_Z, TSM_T, TSM_c, TSM_R, TSM_Q, TSM_a_0, TSM_P_0, Tmask, junk, bj, rj, pj, R1, R2, k; 
struct PV pv0;
struct DS ds0;  ds0=dscreate;

// First, we echo the settings and some descriptive statistics
    m=cols(x);
    ?;"setARNewsNoise: Setting up system matrices for News & Noise model ";
        "    with ";; m;; " vintages and ";; p;; "th order AR dynamics.";
    if      p < 1; "ERROR: p > 0 required! Aborting....";   retp(ds0,pv0); endif;
    
    __altnam = 0 $+ "X_Rev" $+ ftocv(seqa(1,1,m),2,0); __vtype=1;
    ?;call dstat(0,x);?;

// Next, we initialize the control matrices for the state-space model
    TSM_d = zeros(m,1);
    TSM_H = zeros(m,m);
    TSM_c = zeros(2*m+p,1);
    TSM_T = zeros(2*m+p,2*m+p);
    if p>1;
        TSM_T[2:p,1:p-1]=eye(p-1); 
        TSM_Z = ones(m,1)~zeros(m,p-1)~eye(m)~eye(m);
    else;
        TSM_Z = ones(m,1)~eye(m)~eye(m);
    endif;
    TSM_Q = eye(2*m+1);
    TSM_a_0 = zeros(2*m+p,1);
    TSM_P_0 = eye(2*m+p);

//    TSM_a_0[1:p,1] = ones(p,1)*x[1,1];
//    TSM_P_0 = TSM_P_0*1e2; @ This is the SQRT of the P_0 matrix @
//    TSM_P_0 = InitP0(TSM_T,TSM_R,TSM_Q);
//  Echo the resulting model here
//    EchoSSM(TSM_Z,TSM_d,TSM_H,TSM_T,TSM_c,TSM_R,TSM_Q,TSM_a_0,TSM_P_0);

//  Next, we stuff the required matrices into the data structure
    ds0 = reshape(ds0,12,1);
    ds0[1].dataMatrix  = x;
    ds0[2].dataMatrix  = TSM_Z;
    ds0[3].dataMatrix  = TSM_d;
    ds0[4].dataMatrix  = TSM_H;
    ds0[5].dataMatrix  = TSM_c;
    ds0[6].dataMatrix  = TSM_Q;
    ds0[7].dataMatrix  = TSM_a_0;
    ds0[8].dataMatrix  = TSM_P_0;
    ds0[9].dataMatrix  = zeros(m,1); // NsScale
    ds0[10].dataMatrix = zeros(m,1); // NwScale
    ds0[11].dataMatrix = m; 
    ds0[12].dataMatrix = p; 


//  Next, we create the parameter vector
    pv0 = pvCreate;

// Get some initial values by doing an AR(p)
    junk = x[.,m]'~zeros(1,p+2);
    junk = reshape(junk,p+1,rows(x)+p+1)';
    {bj,rj,pj} = olsqr2(junk[.,1],junk[.,2:p+1]);
    TSM_T[1,1:p]= bj';
    Tmask = zeros(2*m+p,2*m+p); Tmask[1,1:p] = ones(1,p);
    pv0 = pvPackm(pv0,TSM_T,"T",Tmask);
    pv0 = pvPack(pv0,stdc(rj),"ARSigma");

//  New initialization code
    R1 = stdc(x[.,2:m]-x[.,1:m-1])|0;
    R2 = R1;
    junk = stdc(x[.,m]);
    for k (1,m-1,1);        
        {bj,rj,pj} = olsqr2(x[.,k+1],x[.,k]);
        if bj <= R1[k,1]^2/junk^2;
            R1[k,1] = 0;
        elseif bj <= 1;
            R1[k,1] = R1[k,1]*sqrt((bj-R1[k,1]^2/junk^2)/(1-R1[k,1]^2/junk^2));
            R2[k,1] = R2[k,1]*sqrt((1-bj)/(1-R2[k,1]^2/junk^2));
        else;
            R2[k,1] = 0;
        endif;            
    endfor;
    pv0 = pvPack(pv0,R1,"RNews");
    pv0 = pvPack(pv0,R2,"RNoise");
//  Old initialization code
//        R1 = stdc(x[.,2:m]-x[.,1:m-1])/sqrt(2);
//        R1 = R1 | stdc(rj)/sqrt(2);
//    pv0 = pvPack(pv0,R1,"RNews");
//        R1 = stdc(x[.,2:m]-x[.,1:m-1])/sqrt(2);
//        R1 = R1 | stdc(rj)/sqrt(2); 
//    pv0 = pvPack(pv0,R1,"RNoise");
    
retp(pv0,ds0);
endp;

@*********************************************************@
@ proc to set up the system matrices for                  @
@  the News + Noise model with AR(p) "true" dynamics      @
@  and first-order Spillover Effects                      @
@*********************************************************@
proc (2) = setARNewsNoiseSpill(x,p);
//
//  Syntax is identical to that for setARNewsNoise(), above.
//  The PV structure is identical as well; only the mask on the T matrix has been changed, 

local m, TSM_T, Tmask, tmp; 
struct PV pv0,pv1;
struct DS ds0;  ds0=dscreate;

// First, set up the model without spillovers
    m=cols(x);
    {pv0,ds0} = setARNewsNoise(x,p);

// Now just change the masking on the T matrix
    "setARNewsNoiseSpill(): Allowing for first-order Spillovers ";
    Tmask = eye(2*m+p); 
    Tmask[1:p,1:p] = zeros(p,p);
    Tmask[1,1:p] = ones(1,p);
    TSM_T = PVunpack(pv0,"T");

//  Stuff it into the new PV structure
    pv1 = pvCreate;
    pv1 = pvPackm(pv1,TSM_T,"T",Tmask);

//  Now just unload the rest of the old structure into the new
    tmp = PVunpack(pv0,"ARSigma");
    pv1 = pvPack(pv1,tmp,"ARSigma");   
    tmp = PVunpack(pv0,"RNews");
    pv1 = pvPack(pv1,tmp,"RNews");  
    tmp = PVunpack(pv0,"RNoise");
    pv1 = pvPack(pv1,tmp,"RNoise");     

retp(pv1,ds0);
endp;

@****************************************************@
@ proc to calculate the llf for                      @
@  the News + Noise model with AR(p) "true" dynamics @
@****************************************************@
proc llfARNewsNoise(struct PV pv1, struct DS ds1, ind);
    local m,p,H,Q,Z,d,T,c,R,a0,P0,y,RNw,RNs,NwScale,NsScale,Qscale,ARSigma,junk;
    
    T   = pvUnpack(pv1,"T");
    y = ds1[1].dataMatrix;
    Z = ds1[2].dataMatrix;
    d = ds1[3].dataMatrix;
    H = ds1[4].dataMatrix; H = H*H';
    c = ds1[5].dataMatrix;
    Q = ds1[6].dataMatrix;
    a0 = ds1[7].dataMatrix;
    // P0 = ds1[8].dataMatrix; P0 = P0*P0';  // Not used - set by InitP0
    NsScale = ds1[9].dataMatrix; 
    NwScale = ds1[10].dataMatrix; 
    m = ds1[11].dataMatrix; 
    p = ds1[12].dataMatrix; 

    R = zeros(2*m+p,2*m+1); 
    R[1,1] = pvUnpack(pv1,"ARSigma");
    RNw = pvUnpack(pv1,"RNews");
    RNs = pvUnpack(pv1,"RNoise");
    if rows(NsScale)>m; // use Qscale instead of NsScale and NwScale
        Qscale=ds1[9].dataMatrix; 
        R[1,1] = R[1,1]*(10^Qscale[1,1]);
        RNw = RNw.*(10^Qscale[2:1+m,1]);
        RNs = RNs.*(10^Qscale[2+m:1+2*m,1]);        
    else;
        RNw = RNw.*(10^NwScale);
        RNs = RNs.*(10^NsScale);
    endif;
    R[1,2:m+1] = -RNw';
//    R[p+1:m+p,2:m+1] = upmat(RNw.*ones(1,m)); // This line was incorrect
    R[p+1:m+p,2:m+1] = upmat(ones(m,1).*RNw'); // corrected version
    R[m+p+1:2*m+p,2+m:2*m+1] = diagrv(eye(m),RNs);


    call SSM_build(Z,d,H,T,c,R,Q,0);
    P0 = InitP0(T,R,Q);
//    EchoSSM(Z,d,H,T,c,R,Q,a0,P0); // Useful for debugging
    call KFiltering(y,a0,P0);

    struct modelResults mm;
    if ind[1];
        mm.function = KF_ml;
    elseif rows(ind) > 3;
        if ind[4] == 1;
            mm.function = KF_matrix(3);
        elseif ind[4] == 2;
            mm.function = KF_matrix(4);
        elseif ind[4] == 3;
            {mm.function,junk} = Ksmoothing;
        elseif ind[4] == 4;
            {junk,mm.function} = Ksmoothing;
        endif;
    endif;
    retp(mm);
endp;

@****************************************************@
@ proc to simulate data for                          @
@  the News + Noise model with AR(p) "true" dynamics @
@****************************************************@
proc GenARNewsNoise(struct PV pv1,n,nburn);
    local m1,p1,RNw,RNs,ARSigma,T1,R1,R2,R3,ytild,yrev,eshocks;
    
    "GenARNewsNoise(): Parameter Vector is:";
    pvgetparnames(pv1)';
    pvgetparvector(pv1)';

    T1   = pvUnpack(pv1,"T");
    ARSigma = pvUnpack(pv1,"ARSigma");
    RNw = pvUnpack(pv1,"RNews");
    RNs = pvUnpack(pv1,"RNoise");

    m1 = rows(RNw); 
    p1 = cols(T1)-m1; 

    R1 = ARSigma | RNw;           
    R2 = -lowmat(RNw.*ones(1,m1));  // Lowmat() since I'm after the *transpose*
    R3 = diagrv(eye(m1),RNs);     

    eshocks = rndn(n+nburn,1+m1);
    ytild = recserar(eshocks*R1,zeros(p1,1),T1[1,1:p1]'); // "True" values
    yrev = ytild + eshocks[.,2:1+m1]*R2 + rndn(n+nburn,m1)*R3;

    "Std. Dev. of vintages:" 
        stdc(yrev[nburn+1:nburn+n,.])';
    "Std. Dev. of measurement errors:" 
        stdc(yrev[nburn+1:nburn+n,.]-ytild[nburn+1:nburn+n,.])';
    "Std. Dev. of revisions:"
        stdc(yrev[nburn+1:nburn+n,2:m1]-yrev[nburn+1:nburn+n,1:m1-1])';

    retp(yrev[nburn+1:nburn+n,.]);
endp;

@****************************************************@
@ proc to calculate the llf for                      @
@  the News + Noise model + Spillover model          @
@ with AR(p) "true" dynamics                         @
@****************************************************@
proc llfARNewsNoiseSpill(struct PV pv1, struct DS ds1, ind);
    local m,p,H,Q,Z,d,T,c,R,a0,P0,y,RNw,RNs,RScale,Tscale,ARSigma,junk;
    
    y = ds1[1].dataMatrix;
    Z = ds1[2].dataMatrix;
    d = ds1[3].dataMatrix;
    H = ds1[4].dataMatrix; H = H*H';
    c = ds1[5].dataMatrix;
    Q = ds1[6].dataMatrix;
    a0 = ds1[7].dataMatrix;
    TScale = ds1[8].dataMatrix;
    RScale = ds1[9].dataMatrix; 
    m = ds1[11].dataMatrix; 
    p = ds1[12].dataMatrix; 

    T = pvUnpack(pv1,"T");
    T[p+1:p+2*m,p+1:p+2*m] = T[p+1:p+2*m,p+1:p+2*m]*diagrv(eye(2*m),10^Tscale);

    R = zeros(2*m+p,2*m+1); 
    R[1,1] = pvUnpack(pv1,"ARSigma");
    RNw = pvUnpack(pv1,"RNews");
    RNs = pvUnpack(pv1,"RNoise");

//    T[1,1:p] = T[1,1:p]*(10^Pscale[1:p,1]);
//    T[p+1:p+2*m,p+1:p+2*m] = T[p+1:p+2*m,p+1:p+2*m]*diagrv(eye(2*m),Pscale[p+1:p+2*m,1]);
//    R[1,1] = R[1,1]*(10^Pscale[p+2*m+1,1]);
//    RNw = RNw.*(10^Pscale[p+2*m+2:p+3*m+1,1]);
//    RNs = RNs.*(10^Pscale[p+3*m+2:p+4*m+1,1]);        
    R[1,1] = R[1,1]*(10^Rscale[1,1]);
    RNw = RNw.*(10^Rscale[2:m+1,1]);
    RNs = RNs.*(10^Rscale[m+2:2*m+1,1]);        

    R[1,2:m+1] = -RNw';
//    R[p+1:m+p,2:m+1] = upmat(RNw.*ones(1,m)); // This line was incorrect
    R[p+1:m+p,2:m+1] = upmat(ones(m,1).*RNw'); // corrected version
    R[m+p+1:2*m+p,2+m:2*m+1] = diagrv(eye(m),RNs);


    call SSM_build(Z,d,H,T,c,R,Q,0);
    P0 = InitP0(T,R,Q);
//    EchoSSM(Z,d,H,T,c,R,Q,a0,P0); // Useful for debugging
    call KFiltering(y,a0,P0);

    struct modelResults mm;
    if ind[1];
        mm.function = KF_ml;
    elseif rows(ind) > 3;
        if ind[4] == 1;
            mm.function = KF_matrix(3);
        elseif ind[4] == 2;
            mm.function = KF_matrix(4);
        elseif ind[4] == 3;
            {mm.function,junk} = Ksmoothing;
        elseif ind[4] == 4;
            {junk,mm.function} = Ksmoothing;
        endif;
    endif;
    retp(mm);
endp;

